# coding=utf-8
from pkg_resources import resource_filename
from unittest import TestCase
from zpretty.cli import choose_prettifier
from zpretty.cli import get_parser
from zpretty.prettifier import ZPrettifier
from zpretty.xml import XMLPrettifier
from zpretty.zcml import ZCMLPrettifier


class TestCli(TestCase):
    """ Test the cli options
    """

    parser = get_parser()

    def test_defaults(self):
        parsed = self.parser.parse_args([])
        self.assertEqual(parsed.file, "-")
        self.assertFalse(parsed.inplace)
        self.assertFalse(parsed.xml)
        self.assertFalse(parsed.zcml)
        self.assertEqual(parsed.encoding, "utf8")

    def test_short_options(self):
        parsed = self.parser.parse_args(["-i", "-x", "-z"])
        self.assertTrue(all((parsed.inplace, parsed.xml, parsed.zcml)))

    def test_long_options(self):
        parsed = self.parser.parse_args(["--inplace", "--xml", "--zcml"])
        self.assertTrue(all((parsed.inplace, parsed.xml, parsed.zcml)))

    def test_file(self):
        html = resource_filename("zpretty.tests", "original/sample_html.html")
        xml = resource_filename("zpretty.tests", "original/sample_xml.xml")
        parsed = self.parser.parse_args([html, xml])
        self.assertEqual(parsed.file, [html, xml])

    def test_choose_prettifier(self):
        """ Check the for the given options and file the best choice is made
        """
        parsed = self.parser.parse_args(["--xml", "--zcml"])
        self.assertEqual(choose_prettifier(parsed, ""), ZCMLPrettifier)
        parsed = self.parser.parse_args(["--xml"])
        self.assertEqual(choose_prettifier(parsed, ""), XMLPrettifier)
        # Without options the extension rules
        parsed = self.parser.parse_args([])
        self.assertEqual(choose_prettifier(parsed, "a.zcml"), ZCMLPrettifier)
        parsed = self.parser.parse_args([])
        self.assertEqual(choose_prettifier(parsed, "a.xml"), XMLPrettifier)
        # The default one is returned if the extension is not recognized
        parsed = self.parser.parse_args([])
        self.assertEqual(choose_prettifier(parsed, "a.txt"), ZPrettifier)



# coding=utf-8
from pkg_resources import resource_filename
from unittest import TestCase
from zpretty.cli import get_parser


class TestReadme(TestCase):
    """ Test zpretty
    """

    def extract_usage_from_readme(self):
        """ Extract the usage from the documentation
        """
        resolved_filename = resource_filename("zpretty", "../README.md")
        readme = open(resolved_filename).read()
        start = readme.index("    zpretty [")
        end = readme.index("\n\nWithout", start)
        return readme[start:end].splitlines()

    def extract_usage_from_parser(self):
        """ Ask the parser for the usage and indent it
        """
        parser = get_parser()
        # temporarily remove the description
        parser.description = None
        parser_help = parser.format_help()
        start = parser_help.index("zpretty [")
        parser_help = parser_help[start:]

        return ["    {0}".format(x).rstrip() for x in parser_help.splitlines()]

    def test_readme(self):
        observed = self.extract_usage_from_readme()
        expected = self.extract_usage_from_parser()
        self.assertListEqual(observed, expected)



# coding=utf-8
from pkg_resources import resource_filename
from unittest import TestCase
from zpretty.zcml import ZCMLPrettifier


class TestZpretty(TestCase):
    """ Test zpretty
    """

    maxDiff = None

    def prettify(self, filename):
        """ Run prettify on filename and check that the output is equal to
        the file content itself
        """
        resolved_filename = resource_filename("zpretty.tests", "original/%s" % filename)
        prettifier = ZCMLPrettifier(resolved_filename)
        observed = prettifier()
        expected = open(resolved_filename).read()
        self.assertListEqual(observed.splitlines(), expected.splitlines())

    def test_zcml(self):
        self.prettify("sample.zcml")



# coding=utf-8
from bs4.builder._lxml import LXMLTreeBuilderForXML
from logging import getLogger
from lxml import etree
from zpretty.attributes import PrettyAttributes
from zpretty.elements import PrettyElement
from zpretty.prettifier import ZPrettifier


logger = getLogger(__name__)


class XMLAttributes(PrettyAttributes):
    """ Customized attribute formatter for zcml
    """

    _valueless_attributes_are_allowed = False
    _known_valueless_attributes = ()
    _multiline_attributes = ()
    _xml_attribute_order = ()

    def sort_attributes(self, name):
        """Sort ZCML attributes in a consistent way
        """
        if name in self._xml_attribute_order:
            return (100 + self._xml_attribute_order.index(name), name)
        return super(XMLAttributes, self).sort_attributes(name)


class XMLElement(PrettyElement):
    attribute_klass = XMLAttributes

    def is_self_closing(self):
        """ Is this element self closing?
        """
        if not self.is_tag():
            raise ValueError("This is not a tag")
        # Just check if the element has some content.
        return not self.getchildren()

    @property
    def tag(self):
        """ Return the tag name
        """
        if not self.context.prefix:
            return self.context.name
        else:
            return u":".join((self.context.prefix, self.context.name))

    def render_text(self):
        """ Add an empty line between each element
        """
        return super(XMLElement, self).render_text()


class XMLTreeBuilder(LXMLTreeBuilderForXML):
    """ Override the default Tree builder
    """

    def default_parser(self, encoding):
        # This can either return a parser object or a class, which
        # will be instantiated with default arguments.
        if self._default_parser is not None:
            return self._default_parser
        return etree.XMLParser(
            target=self,
            strip_cdata=False,
            recover=True,
            encoding=encoding,
            remove_blank_text=False,
            attribute_defaults=False,
            dtd_validation=False,
            load_dtd=False,
            no_network=True,
            ns_clean=True,
            resolve_entities=False,
            remove_comments=False,
            remove_pis=False,
            collect_ids=False,
            compact=False,
        )


class XMLPrettifier(ZPrettifier):
    """ Prettify according to the ZCML style guide
    """

    parser = "xml"
    pretty_element = XMLElement
    builder = XMLTreeBuilder()



# coding=utf-8
from logging import getLogger


try:
    from html import escape
except ImportError:
    # Python < 3.8
    from cgi import escape


logger = getLogger(__name__)


class PrettyAttributes(object):
    """ Render attributes in a pretty way.

    - one per line
    - sorted semantically and alphabetically
    - with properly indented and escaped values
    """

    _attribute_template = u'%s="%s"'
    _valueless_attributes_are_allowed = True
    _known_valueless_attributes = ("hidden", "required")
    _multiline_prefix = u"  "
    _multiline_attributes = ()
    _tal_multiline_attributes = (
        u"attributes",
        u"define",
        u"tal:attributes",
        u"tal:define",
    )

    _tal_attribute_order = (
        "tal:define",
        "tal:switch",
        "tal:condition",
        "tal:repeat",
        "tal:case",
        "tal:content",
        "tal:replace",
        "tal:omit-tag",
        "tal:attributes",
        "tal:on-error",
    )

    _i18n_attributes = (
        "i18n:translate",
        "i18n:domain",
        "i18n:context",
        "i18n:source",
        "i18n:target",
        "i18n:name",
        "i18n:attributes",
        "i18n:data",
        "i18n:comment",
        "i18n:ignore",
        "i18n:ignore-attributes",
    )

    def __init__(self, attributes, element=None):
        """ attributes is a dict like object
        """
        self.attributes = attributes
        self.element = element

    def sort_attributes(self, name):
        """This sorts the attribute trying to group them semantically

        Starting from the top:

        1. xml namespaces
        2. class, id
        3. attributes not belonging in to other categories (default)
        4. data- attributes
        5. tal attributes
        6. i18n attributes
        """
        if name.startswith("xmlns"):
            return (0, name)
        if name in ("class", "id"):
            return (100, name)
        if name.startswith("data"):
            return (300, name)
        if "tal:" + name in self._tal_attribute_order:
            tal_index = self._tal_attribute_order.index("tal:" + name)
            return (400 + tal_index, name)
        if name in self._tal_attribute_order:
            tal_index = self._tal_attribute_order.index(name)
            return (400 + tal_index, name)
        if name in self._i18n_attributes:
            return (900, name)
        return (200, name)

    def format_multiline(self, name, value):
        """
        """
        value_lines = filter(None, value.split())
        line_joiner = u"\n" + (u" " * (len(name) + 2))
        return line_joiner.join(value_lines)

    def format_tal_multiline(self, value):
        """ There are some tal specific attributes that contain ; separated
        statements.
        They are used to define variables or set other attributes.
        You can define many variables by adding statements separated by ';'.
        If the statement contains a ';', it will be escaped as ';;'.

        It is convenient to always have those attribute in the form:

        tal:define="
          var1 statement1;
          var2 statement2;
          ...
        "
        """
        # temp skip ';;' the escape sequence to enter a ';' in a statement
        statements = value.replace(";;", "<>").split(";")

        # We always want an empty line first...
        lines = [u""]
        for statement in statements:
            statement = statement.strip()
            if statement:
                if not statement.endswith(";"):
                    statement += u";"
                lines.append(self._multiline_prefix + statement)
        # ... and at the end
        lines.append(u"")

        new_value = u"\n".join(lines)
        # restore ';;'
        return new_value.replace("<>", ";;")

    def can_be_valueless(self, name):
        """ Check if the attribute name can be without a value
        """
        if not self._valueless_attributes_are_allowed:
            return False
        if name.startswith("data-"):
            return True
        if name in self._known_valueless_attributes:
            return True
        return False

    def lines(self):
        """ Take the attributes, sort them and prettify their values
        """
        attributes = self.attributes
        sorted_names = sorted(attributes, key=self.sort_attributes)
        lines = []
        for name in sorted_names:
            value = attributes[name]
            if isinstance(value, list):
                # Happens, e.g., for the class attribute
                value = u" ".join(value)
            if name in self._multiline_attributes:
                value = self.format_multiline(name, value)
            elif name in self._tal_multiline_attributes:
                value = self.format_tal_multiline(value)
            if not value and self.can_be_valueless(name):
                line = name
            else:
                line = self._attribute_template % (name, escape(value))
            lines.append(line)
        return lines

    def __call__(self):
        """ Render the attributes as text
        """
        return u"\n".join(self.lines())



# coding=utf-8
from pkg_resources import resource_filename
from unittest import TestCase
from zpretty.prettifier import ZPrettifier


class TestZpretty(TestCase):
    """ Test zpretty
    """

    maxDiff = None

    def assertPrettified(self, original, expected, encoding="utf8"):
        """ Check if the original html has been prettified as expected
        """
        if isinstance(expected, tuple):
            expected = u"\n".join(expected)
        prettifier = ZPrettifier(text=original, encoding=encoding)
        observed = prettifier()
        self.assertEqual(observed, expected)

    def prettify(self, filename):
        """ Run prettify on filename and check that the output is equal to
        the file content itself
        """
        resolved_filename = resource_filename("zpretty.tests", "original/%s" % filename)
        prettifier = ZPrettifier(resolved_filename)
        observed = prettifier()
        expected = open(resolved_filename).read()
        self.assertListEqual(observed.splitlines(), expected.splitlines())

    def test_format_self_closing_tag(self):
        self.assertPrettified("<tal:test />", u"<tal:test />\n")

    def test_nesting_no_text(self):
        # no attributes
        self.assertPrettified(
            "<root><tal:test /></root>", u"<root><tal:test /></root>\n"
        )

        self.assertPrettified(
            "<root><tal:test /> </root>", u"<root><tal:test />\n</root>\n"
        )

        self.assertPrettified(
            "<root> <tal:test /> </root>", u"<root>\n  <tal:test />\n</root>\n"
        )

        self.assertPrettified(
            "<root> <tal:test /></root>", u"<root>\n  <tal:test /></root>\n"
        )
        self.assertPrettified(
            "<root><tal:test><child /></tal:test></root>",
            u"<root><tal:test><child></child></tal:test></root>\n",
        )

    def test_nesting_with_text(self):
        # no attributes
        self.assertPrettified(
            "<root> a<tal:test /></root>", u"<root>\n  a<tal:test /></root>\n"
        )

        self.assertPrettified(
            "<root>a <tal:test /> </root>", u"<root>a\n  <tal:test />\n</root>\n"
        )

        self.assertPrettified(
            "<root> a <tal:test /> </root>", u"<root>\n  a\n  <tal:test />\n</root>\n"
        )

        self.assertPrettified(
            "<root> a <div /> </root>", u"<root>\n  a\n  <div></div>\n</root>\n"
        )
        self.assertPrettified(
            "<root> a <div /> <div /> </root>",
            u"<root>\n  a\n  <div></div>\n  <div></div>\n</root>\n",
        )
        self.assertPrettified(
            "<root> a <div />\n <div /> </root>",
            u"<root>\n  a\n  <div></div>\n  <div></div>\n</root>\n",
        )
        self.assertPrettified(
            "<root> a <div />\na <div /> </root>",
            u"<root>\n  a\n  <div></div>\na\n  <div></div>\n</root>\n",
        )
        self.assertPrettified(
            "<root> <div>a</div> </root>", u"<root>\n  <div>a</div>\n</root>\n"
        )
        self.assertPrettified(
            "<root> <div>a\n  </div> </root>", u"<root>\n  <div>a\n  </div>\n</root>\n"
        )
        self.assertPrettified("<div><p>a</p></div>", u"<div><p>a</p></div>\n")

    def test_nesting_with_tail(self):
        # no attributes
        self.assertPrettified(
            "<root><p></p><!-- #a -->\n\n<p></p></root>",
            u"<root><p></p><!-- #a -->\n\n  <p></p></root>\n",
        )
        self.assertPrettified(
            "<root><tal:test />a</root>", u"<root><tal:test />a</root>\n"
        )
        self.assertPrettified(
            "<root><tal:test />a </root>", (u"<root><tal:test />a", u"</root>", u"")
        )

        self.assertPrettified(
            "<root><tal:test /> a </root>",
            (u"<root><tal:test />", u"  a", u"</root>", u""),
        )
        self.assertPrettified(
            "<root><tal:test /> a </root>", u"<root><tal:test />\n  a\n</root>\n"
        )

    def test_many_children(self):
        """
        """
        self.assertPrettified(
            u"<root><tal:test /><div></div></root>",
            u"<root><tal:test /><div></div></root>\n",
        )
        self.assertPrettified(
            u"<root><tal:test /> <div></div></root>",
            u"<root><tal:test />\n  <div></div></root>\n",
        )

    def test_valueless_attributes(self):
        """ Test attributes without value
        (hidden, required, data-attributes, ...)
        Some of them are rendered valueless, some other not.
        """
        self.assertPrettified(
            u'<root data-attribute=""></root>', u"<root data-attribute></root>\n"
        )
        self.assertPrettified(u"<root hidden></root>", u"<root hidden></root>\n")
        self.assertPrettified(u"<root selected></root>", u'<root selected=""></root>\n')

    def test_fix_self_closing(self):
        """ Check if open self closing tags are rendered correctly
        """
        self.assertPrettified(u"<input><img><input>", u"<input /><img /><input />\n")
        self.assertPrettified(u"<input><a /><b />", u"<input /><a></a><b></b>\n")
        self.assertPrettified(
            u"<input><a /><b /></input>", u"<input /><a></a><b></b>\n"
        )

    def test_element_repr(self):
        prettifier = ZPrettifier(text="")
        self.assertEqual(repr(prettifier.root), "<pretty:-1:null_tag_name />")

    def test_whitelines_not_stripped(self):
        self.assertPrettified(u"<root>\n</root>", u"<root>\n</root>\n")
        self.assertPrettified(
            u"<root>\n    Hello!   \n</root>", u"<root>\n    Hello!\n</root>\n"
        )

    def test_entities(self):
        self.assertPrettified("<root>&nbsp;</root>", u"<root>&nbsp;</root>\n")

    def test_sample_html(self):
        self.prettify("sample_html.html")

    def test_text_with_markup(self):
        self.prettify("text_with_markup.md")



# coding=utf-8
from unittest import TestCase
from zpretty.attributes import PrettyAttributes


class TestZPrettyAttributess(TestCase):
    """ Test zpretty
    """

    def assertPrettifiedAttributes(self, attributes, expected, level=0):
        """ Check if the attributes are properly sorted and formatted
        """
        pretty_attribute = PrettyAttributes(attributes)
        observed = pretty_attribute()
        self.assertEqual(observed, expected)

    def test_no_attributes(self):
        self.assertPrettifiedAttributes({}, u"")

    def test_one_attribute(self):
        self.assertPrettifiedAttributes({u"a": u"1"}, u'a="1"')

    def test_tal_define(self):
        self.assertPrettifiedAttributes(
            {u"tal:define": u"a 1; b 2"},
            u"\n".join((u'tal:define="', u"  a 1;", u"  b 2;", u'"')),
        )

    def test_format_attributes_many_attribute(self):
        self.assertPrettifiedAttributes(
            {u"a": u"1", u"b": u"2", u"class": u"hidden", u"tal:define": u"a 1; b 2"},
            u"\n".join(
                (
                    u'class="hidden"',
                    u'a="1"',
                    u'b="2"',
                    u'tal:define="',
                    u"  a 1;",
                    u"  b 2;",
                    u'"',
                )
            ),
        )



# coding=utf-8
from bs4 import BeautifulSoup
from logging import getLogger
from uuid import uuid4
from zpretty.elements import PrettyElement

import fileinput


logger = getLogger(__name__)


class ZPrettifier(object):
    """ Wraps and renders some text that may contain xml like stuff
    """

    pretty_element = PrettyElement
    parser = "html.parser"
    builder = None
    _end_with_newline = True
    _newlines_marker = str(uuid4())

    def __init__(self, filename="", text="", encoding="utf8"):
        """ Create a prettifier instance taking the contents
        from a text or a filename
        """
        self.encoding = encoding
        self.filename = filename
        if self.filename:
            text = "".join(fileinput.input([filename]))
        if not isinstance(text, str):
            text = text.decode(self.encoding)
        self.text = "\n".join(
            line if line.strip() else self._newlines_marker
            for line in text.splitlines()
        )
        self.soup = self.get_soup(self.text)
        self.root = self.pretty_element(self.soup, -1)

    def get_soup(self, text):
        """ Tries to get the soup from the given test

        At first it will try to parse the text:

        1. as an xml
        2. as an html
        3. will just return the unparsed text
        """
        markup = u"<{null}>{text}</{null}>".format(
            null=self.pretty_element.null_tag_name, text=text
        )
        wrapped_soup = BeautifulSoup(markup, self.parser, builder=self.builder)
        return getattr(wrapped_soup, self.pretty_element.null_tag_name)

    def autofix(self):
        """ Do various autofx on the soup
        """

    def pretty_print(self, el):
        """ Pretty print an element indenting it based on level
        """
        prettified = el().replace(self._newlines_marker, u"")
        if self._end_with_newline and not prettified.endswith(u"\n"):
            prettified += u"\n"
        return prettified

    def __call__(self):
        self.autofix()
        return self.pretty_print(self.root)



# coding=utf-8
from logging import getLogger
from zpretty.xml import XMLAttributes
from zpretty.xml import XMLElement
from zpretty.xml import XMLPrettifier


logger = getLogger(__name__)


class ZCMLAttributes(XMLAttributes):
    """ Customized attribute formatter for zcml
    """

    _multiline_attributes = ("for",)
    _xml_attribute_order = (
        "name",
        "title",
        "description",
        "package",
        "file",
        "provides",
        "for",
        "factory",
        "manager",
        "permission",
        "class",
        "allowed_attributes",
        "attribute",
        "template",
        "layer",
    )


class ZCMLElement(XMLElement):
    first_attribute_on_new_line = True
    before_closing_multiline = u"    "
    attribute_klass = ZCMLAttributes

    def render_text(self):
        """ Add an empty line between each element
        """
        return super(ZCMLElement, self).render_text()


class ZCMLPrettifier(XMLPrettifier):
    """ Prettify according to the ZCML style guide
    """

    pretty_element = ZCMLElement



# coding=utf-8
from argparse import ArgumentParser
from os.path import splitext
from sys import stdout
from zpretty.prettifier import ZPrettifier
from zpretty.xml import XMLPrettifier
from zpretty.zcml import ZCMLPrettifier


def get_parser():
    parser = ArgumentParser(
        prog="zpretty",
        description="An opinionated HTML/XML soup formatter",
        epilog=None,
    )
    parser.add_argument(
        "--encoding",
        help="The file encoding (defaults to utf8)",
        action="store",
        dest="encoding",
        default="utf8",
    )
    parser.add_argument(
        "-i",
        "--inplace",
        help="Format files in place (overwrite existing file)",
        action="store_true",
        dest="inplace",
        default=False,
    )
    parser.add_argument(
        "-x",
        "--xml",
        help="Threat the input file(s) as XML",
        action="store_true",
        dest="xml",
        default=False,
    )
    parser.add_argument(
        "-z",
        "--zcml",
        help="Threat the input file(s) as XML. Follow the ZCML styleguide",
        action="store_true",
        dest="zcml",
        default=False,
    )
    parser.add_argument(
        "file",
        nargs="*",
        default="-",
        help="The list of files to prettify (defaults to stdin)",
    )
    return parser


def choose_prettifier(config, infile):
    """ Choose the best prettifier given the config and the input file
    """
    if config.zcml:
        return ZCMLPrettifier
    if config.xml:
        return XMLPrettifier
    ext = splitext(infile)[-1].lower()
    if ext == ".xml":
        return XMLPrettifier
    if ext == ".zcml":
        return ZCMLPrettifier
    return ZPrettifier


def run():
    """ Prettify each filename passed in the command line
    """
    parser = get_parser()
    config = parser.parse_args()
    encoding = config.encoding
    for infile in config.file:
        Prettifier = choose_prettifier(config, infile)
        prettifier = Prettifier(infile, encoding=encoding)
        prettified = prettifier().encode(encoding)
        if config.inplace and not infile == "-":
            with open(infile, "wb") as f:
                f.write(prettified)
        else:
            stdout.write(prettified)


if __name__ == "__main__":
    run()
